/*
  Calisto - An internet Talker
  Copyright (C) 1998-2001 Peter Howkins, Matthew Howkins and Simon Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,     
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License   
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: calisto@calisto.cx

  $Id: commands.c,v 1.57 2002/12/02 19:37:04 peter Exp $

 */
static const char rcsid[] = "$Id: commands.c,v 1.57 2002/12/02 19:37:04 peter Exp $";

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "unistd.h"

#include "config.h"
#include "colours.h"
#include "commands.h"
#include "globals.h"
#include "help.h"
#include "library.h"
#include "log.h"
#include "mem.h"
#include "msnprintf.h"
#include "playerdb.h"
#ifdef USE_PLUGINS
 #include "plugin2.h"
#endif /* USE_PLUGINS */
#include "privs.h"
#include "socks.h"
#include "structs.h"
#include "strplus.h"

typedef struct {
  char *name;
  char *shortcut;
  void (*function)(character *, char *, const char *, void *);
  char *privname; 
  size_t offset;
} commandl;

typedef struct {
  commandl c;
  listnode commandlink;
} command_list_item;

void command_a(character *, char *, const char *, void *);
void command_bug(character *, char *, const char *, void *);
void command_cls(character *, char *, const char *, void *);
void command_commands(character *, char *, const char *, void *);
void command_dc(character *, char *, const char *, void *);
void command_finger(character *, char *, const char *, void *);
void command_grant(character *, char *, const char *, void *);
void command_group(character *, char *, const char *, void *);
void command_invite(character *, char *, const char *, void *);
void command_lsd(character *, char *, const char *, void *);
void command_motd(character *, char *, const char *, void *);
void command_ping(character *, char *, const char *, void *);
void command_prompt(character *, char *, const char *, void *);
void command_quit(character *, char *, const char *, void *);
void command_recap(character *, char *, const char *, void *);
void command_reload(character *, char *, const char *, void *);
void command_resident(character *, char *, const char *, void *);
void command_remove(character *, char *, const char *, void *);
void command_save(character *, char *, const char *, void *);
void command_say(character *, char *, const char *, void *);
void command_set(character *, char *, const char *, void *);
void command_setrank(character *, char *, const char *, void *);
void command_shout(character *, char *, const char *, void *);
void command_shutdown(character *, char *, const char *, void *);
void command_stat(character *, char *, const char *, void *);
void command_tell(character *, char *, const char *, void *);
void command_time(character *, char *, const char *, void *);
void command_toggle(character *, char *, const char *, void *);
void command_users(character *, char *, const char *, void *);
void command_version(character *, char *, const char *, void *);
void command_who(character *, char *, const char *, void *);
void command_whoami(character *, char *, const char *, void *);

const command commands[] = {
  { "a",     "",  command_a,      "n_base" },
  { "bug",     "",  command_bug,      "n_base" },
  { "bump",    "",  command_dc,       "s_bump" },
  { "clear",   "",  command_cls,      "n_base" },
  { "cls",     "",  command_cls,      "n_base" },
  { "commands","",  command_commands, "n_base" },
  { "dc",      "",  command_dc,       "s_bump" },
  { "emote",   ";:",command_say,      "n_talk" },
  { "exit",    "",  command_quit,     "n_base" },
  { "finger",  "",  command_finger,   "n_base" },
  { "grant",   "",  command_grant,    "a_grant" },
  { "group",   "",  command_group,    "n_base" },
  { "help",    "",  command_help,     "n_base" }, /* Found in help.c */
  { "idea",    "",  command_bug,      "n_base" },
  { "invite",  "",  command_invite,   "n_base" },
  { "lsd",     "",  command_lsd,      "s_trace" },
  { "motd",    "",  command_motd,     "n_base" },
  { "ping",    "",  command_ping,     "n_base" },
  { "prompt",  "",  command_prompt,   "n_base" },
  { "quit",    "",  command_quit,     "n_base" },
  { "recap",   "",  command_recap,    "n_base" },
  { "reload",  "",  command_reload,   "a_base" },
  { "remote",  ",", command_tell,     "n_talk" },
  { "remove",  "",  command_remove,   "a_grant" },
  { "res",     "",  command_resident, "s_base" },
  { "resident","",  command_resident, "s_base" },
  { "save",    "",  command_save,     "n_save" },
  { "say",     "'", command_say,      "n_talk" },
  { "set",     "",  command_set,      "a_base" },
  { "setrank", "",  command_setrank,  "a_grant" },
  { "shout",   "",  command_shout,    "n_talk" },
  { "shutdown","",  command_shutdown, "a_shutdown" },
  { "stat",    "",  command_stat,     "a_base" },
  { "sumotd",  "",  command_motd,     "s_base" },
  { "tell",    ".", command_tell,     "n_talk" },
  { "time",    "",  command_time,     "n_base" },
  { "toggle",  "",  command_toggle,   "n_base" },
  { "typo",    "",  command_bug,      "n_base" },
  { "users",   "",  command_users,    "n_base" },
  { "version", "",  command_version,  "n_base" },
  { "who",     "",  command_who,      "n_base" },
  { "whoami",  "",  command_whoami,   "n_base" },
  { "x",       "",  command_finger,   "n_base" }
};
const unsigned commandn = sizeof commands / sizeof(command);

pool *command_pool = 0;
listhead AllCommands;

/* the compare function used to sort the command linked list */
static int compare_command(const listnode *a, const listnode *b)
{
  const command_list_item *c =
    LIST_GET_DATA(a, command_list_item *, commandlink);
  const command_list_item *d =
    LIST_GET_DATA(b, command_list_item *, commandlink);

  return strcmp(c->c.name, d->c.name);
}

void init_commands(void)
{
  unsigned int c;
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */
  /* initialise commands linked list */
  LIST_INIT(AllCommands);

  /* Initialise pool-based memory management */
  command_pool = pool_create(sizeof(command_list_item));
  if (!command_pool) {
    fprintf(stderr, "Commands: Not enough memory creating command_pool\n");
    log(usage, "Commands: Not enough memory creating command_pool");
    exit(EXIT_FAILURE);
  }

  /* setup the core commands */
  for (c = 0; c < commandn; c++) {
    command_list_item *command_new = pool_malloc(command_pool);
    if(!command_new) {
      fprintf(stderr, "init_commands: Out of memory.");
      log(usage, "init_commands: Out of memory");
      exit(sizeof(command_list_item));
    }

    command_new->c.name = commands[c].name;
    command_new->c.shortcut = commands[c].shortcut;
    command_new->c.function = commands[c].function;
    command_new->c.privname = commands[c].privname;
    command_new->c.offset   = NO_PLUGIN_EXTRA_DATA;

    LIST_LINK_NODE_AT_END(AllCommands, &(command_new->commandlink));
  }

#ifdef USE_PLUGINS
  /* setup the plugins commands */
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    if(p->clist != NULL) {
      command *c = p->clist;
      while(c->name != NULL) {
        command_list_item *command_new = pool_malloc(command_pool);
        if(!command_new) {
          fprintf(stderr, "init_commands: Out of memory.");
          log(usage, "init_commands: Out of memory");
          exit(sizeof(command_list_item));
        }

        command_new->c.name = c->name;
        command_new->c.shortcut = c->shortcut;
        command_new->c.function = c->function;
        command_new->c.privname = c->privname;
        command_new->c.offset = p->offset;

        LIST_LINK_NODE_AT_END(AllCommands, &(command_new->commandlink));
        c++;
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  /* Sort the commands in alphabetical order */
  list_qsort(&AllCommands, compare_command); 

}

/* free up memory created by commands, and destruct the list */
void free_commands(void)
{
  listnode *cNode = AllCommands.head.next;

  while(LIST_NODE_IS_REAL(cNode)) {
    command_list_item *c = LIST_GET_DATA(cNode, command_list_item *, 
                                         commandlink);
    LIST_UNLINK_NODE(cNode);
    pool_free(c);
    cNode = cNode->next;
  }
  if(!LIST_IS_EMPTY(AllCommands))
    log(debug, "free_commands: error emptying Commands list");

  pool_destroy(command_pool);
}

void command_version(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c);
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */

  bar2(des, "Version", NULL, NULL);
  send_to_char(c, "^HCalisto^n %u.%02u GPL (c) 1998-2001 ^HFlibble^n, ^HThe_Pope^n and ^HWilfred^n\n", version/100, version%100);
  send_to_char(c, "  ^H+^n Colour Code - ^Hekto^n (ekto@ekto.org)\n");
  send_to_char(c, "  ^H+^n snprintf    - based on code by ^HAlain Magloire^n (alainm@rcsm.ee.mcgill.ca)\n");

#ifdef USE_PLUGINS
  /* Plugins */
  if(LIST_NODE_IS_REAL(pNode))
    send_to_char(c, "Active Plugins\n");

  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    send_to_char(c, "  ^H+^n %-16 16s v%u.%02u - ^H%s^n\n", p->name, p->version/100, p->version%100, p->author);
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  bar2(des, NULL, NULL, NULL);
}

void command_toggle(character *c, char *r, const char *comm, void *s)
{
  char *tog;

  if(STREQ(r, "")) {
    help_usage(c, comm);
    return;
  }

  strmunch(r, &tog, &r);

  if(STREQ(tog, "")) {
    help_usage(c, comm);
    return;
  }
  /* option must be on or off */
  if(!STRIEQ(r, "on") && !STRIEQ(r, "off")) {
    send_to_char(c, "Toggle must be on or off\n");
    help_usage(c, comm);
    return;
  }

  if(!prefname_exists(tog)) {
    send_to_char(c, "preference not found\n");
    help_usage(c, comm);
    return;
  }

  if(STRIEQ(r, "on")) {
    setpref(c, strlower(tog));
  } else {
    clearpref(c, strlower(tog));
  }


}

void command_invite(character *c, char *r, const char *comm, void *s)
{
  /* **** minor rewrite required when 'ranks' are added */
  char path[256];
  FILE *fp;

  if(talker_level != TALKER_INVITE && talker_level != TALKER_SU_INVITE) {
    send_to_char(c, "This command has no use in the talkers current mode\n");
    return;
  }

  /* this is not nice, we ought to insert into the commands linked list with
     the correct priv */
  if(talker_level == TALKER_SU_INVITE && !haspriv(c, "s_base")) {
    send_to_char(c, "This command is for staff only in the talkers current mode\n");
    return;
  }

  if(STREQ(r, "")) {
    help_usage(c, comm);
    return;
  }

  /* valid name check */
  if(validate_name(r) == FALSE) {
    send_to_char(c,
     "Names must be between %d and %d characters long and only contain the\n"
     "characters (a-z, A-Z, 0-9)\n", MIN_NAME_LENGTH, MAX_NAME_LENGTH);
    return;
  }
  
  /* are they allready a player ? */
  if(player_exists(r)) {
    send_to_char(c, "That player is allready a resident\n");
    return;
  }

  /* are they allready invited ? */
  msnprintf(path, sizeof(path), "_%s", r);
  if(player_exists(path)) {
    send_to_char(c, "Someone has already invited that player\n");
    return;
  }

  /* invite player */
  msnprintf(path, sizeof(path), "%s_%s", PLAYER_DIR, r);
  fp = fopen(path, "w");
  if(!fp) {
    send_to_char(c, "Error inviting player\n");
    log(usage, "Error inviting player, falied to open '%s'", path);
    return;
  }
  fprintf(fp, "player invited by '%s'", c->name);
  fclose(fp);

  send_to_char(c, "Player %s invited\n", r);

}

void command_resident(character *c, char *r, const char *comm, void *s)
{
  /* **** minor rewrite required when 'ranks' are added */

  character *vict;

  if(talker_level != TALKER_RESIDENCY) {
    send_to_char(c, "This command has no use in the talkers current mode\n");
    return;
  }
  if(STREQ(r, "")) {
    help_usage(c, comm);
    return;
  }
  
  vict = character_from_name(r);
  if(!vict) {
    send_to_char(c, "Player '%s' not found or ambigous\n");
  } else {
    /* this doesn't prevent someone who is a resident being made a resident
       again */
    setpriv(vict, "resident");
    save_player(getdes(vict), vict->name);
    send_to_char(vict, "%s has made you a resident, you character will now be saved\n", c->name);
    send_to_char(c, "%s is now a resident\n", vict->name);
  }
}

/* return a string of either the hostname or the IP address */
char *format_host(char *buf, size_t bufsize, descriptor *des)
{
  if(STREQ(des->hostname, "") || (use_net_lookups == 0)) { 
#ifdef IPV6
    STRNCOPY(buf, des->hostname, bufsize);
#else
    ip4_to_string(buf, bufsize, &des->ip_addr);
#endif
  } else {
    STRNCOPY(buf, des->hostname, bufsize);
  }
  return buf;
}

void command_a(character *c, char *r, const char *comm, void *s)
{
  listnode *cNode = AllCommands.head.next;

  while(LIST_NODE_IS_REAL(cNode)) {
    command_list_item *cli=LIST_GET_DATA(cNode, command_list_item *, commandlink); 
    commandl *co = &cli->c;

     send_to_char(c, "%s - %s - %s - %d\n", co->name, co->shortcut,
                                            co->privname, co->offset);

     cNode =cNode->next;
   }
}

void command_set(character *c, char *r, const char *comm, void *s)
{
  char *variable;
  char *value;
  bool quoted = FALSE;

  if (STREQ(r, "")) {
    send_to_char(c, "use_net_lookups: %d\n", use_net_lookups);
    if(idle_boot_time == 0) {
      send_to_char(c, "idle_boot_time:  Disabled (0)\n");
    } else {
      send_to_char(c, "idle_boot_time:  %u minutes\n", idle_boot_time);
    }
    send_to_char(c, "max_connections: %u\n", max_connections);
    return;
  }

  /* Break input into variable and value */
  strmunch(r, &variable, &r);
  if (*r == '\"') {
    quoted = TRUE;
    r++;
  }
  value = r;
  if (quoted) {
    while (*r && *r != '\"')
      r++;
  } else {
    while (*r && !isspace(*r))
      r++;
  }
  *r = '\0';
/*  send_to_char(c, "variable = \'%s\'\n", variable);
  send_to_char(c, "value    = \'%s\'\n", value);*/

  /* Check if value specified OK */
  if (STREQ(value, "")) {
    help_usage(c, "set");
    return;
  }

  /* Handle known variables, and complain about unknown ones */
  if (STRIEQ(variable, "use_net_lookups")) {
    use_net_lookups = atoi(value);
    if (use_net_lookups)
      use_net_lookups = TRUE;
  } else if (STRIEQ(variable, "idle_boot_time")) {
    idle_boot_time = (unsigned) atoi(value);
  } else if (STRIEQ(variable, "max_connections")) {
    max_connections = (unsigned) atoi(value);
  } else {
    send_to_char(c, "Unknown variable '%s'\n"
                 "Use the set command without parameters to list variables.\n",
                 variable);
  }
}

void command_recap(character *c, char *r, const char *comm, void *s)
{
  if(STREQ(r, "")) {
    help_usage(c, "recap");
    return; 
  }
  if(STRIEQ(r, c->name)) {
    STRNCOPY(c->name, r, sizeof(c->name));
    send_to_char(c, "Name recapitalised\n");
  } else {
    send_to_char(c, "Input must be the same as your name\n");
  }
}

void command_save(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c);
  int error = save_player(des, c->name);
  if(error) {
    send_to_char(c, "Error saving your character\n");
  } else {
    send_to_char(c, "Successfully saved your character\n");
  }
}


void command_cls(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c);
  int i;

  for (i = 0; i < des->term_height; i++) {
    send_to_char(c, "\n");
  }
/*  send_to_char(c, "\x1b[0;0H"); */
  send_to_char(c, "\x1b[%dA", des->term_height);
}

void command_reload(character *c, char *r, const char *comm, void *s)
{
  if (STRIEQ(r, "help")) {
    help_reload(c);
#ifdef USE_PLUGINS
  } else if (STRIEQ(r, "plugins")) {
    plugin_reload(c);
#endif /* USE_PLUGINS */
  } else if (STRIEQ(r, "all")) {
    help_reload(c);
  } else {
    send_to_char(c, "Unrecognised reload option %s\n", r);
  }
}

#define COMMAND_STAT_OPTIONS "net mem"
static void command_stat_usage(character *c)
{
  send_to_char(c, "Usage: stat <option> [<option>...]\n");
  send_to_char(c, "where <option> is one of `%s all`\n",
               COMMAND_STAT_OPTIONS);
}

void command_stat(character *c, char *r, const char *comm, void *s)
{
  if (*r) {
    while (*r) {
      char *sub;

      strmunch(r, &sub, &r);
      if (STRIEQ(sub, "mem")) {
        pool_stat_t ps  = pool_stat(descriptor_pool);
        pool_stat_t ps2 = pool_stat(command_pool);
#ifdef USE_PLUGINS
        pool_stat_t ps3 = pool_stat(plugin_pool);
#endif /* USE_PLUGINS */
        send_to_char(c, "^H[Memory - Descriptors]^n\n");
        send_to_char(c, "   Block Size : %u\n", ps.block_size);
        send_to_char(c, "        Usage : %u\n", ps.usage);
        send_to_char(c, "     Capacity : %u (%.1fk)\n", ps.capacity,
                     (double) (ps.block_size * ps.capacity) / 1024.0);
        send_to_char(c, "         Free : %u\n", ps.capacity - ps.usage);
        send_to_char(c, "  Connections : %u/%u\n", ps.usage,
                     max_connections);


        send_to_char(c, "^H[Memory - Commands]^n\n");
        send_to_char(c, "   Block Size : %u\n", ps2.block_size);
        send_to_char(c, "        Usage : %u\n", ps2.usage);
        send_to_char(c, "     Capacity : %u (%.1fk)\n", ps2.capacity,
                     (double) (ps2.block_size * ps2.capacity) / 1024.0);
        send_to_char(c, "         Free : %u\n", ps2.capacity - ps2.usage);
#ifdef USE_PLUGINS
        send_to_char(c, "^H[Memory - Plugins]^n\n");
        send_to_char(c, "   Block Size : %u\n", ps3.block_size);
        send_to_char(c, "        Usage : %u\n", ps3.usage);
        send_to_char(c, "     Capacity : %u (%.1fk)\n", ps3.capacity,
                     (double) (ps3.block_size * ps3.capacity) / 1024.0);
        send_to_char(c, "         Free : %u\n", ps3.capacity - ps3.usage);
#endif /* USE_PLUGINS */
      } else if (STRIEQ(sub, "net")) {
        FILE *f;
        const char *hostname = getenv("HOSTNAME");
        const char *machtype = getenv("MACHTYPE");

        send_to_char(c, "^H[Network and Machine]^n\n");
        send_to_char(c, "    PID : %u\n", (unsigned) getpid());
        send_to_char(c, "   PPID : %u\n", (unsigned) getppid());
        send_to_char(c, "   HOST : %s\n", hostname ? hostname : "** Unknown **");
        send_to_char(c, "   TYPE : %s\n", machtype ? machtype : "** Unknown **");
        /* Display load average of server, if available */
        f = fopen("/proc/loadavg", "r");
        if (f) {
          char buffer[256];
          float load1, load5, load15;

          if (fgets(buffer, sizeof(buffer), f)) {
            if (sscanf(buffer, "%f %f %f", &load1, &load5, &load15) == 3) {
              send_to_char(c, "LoadAvg : %.2f %.2f %.2f\n",
                           load1, load5, load15);
            }
          }
          fclose(f);
        }
      } else if (STRIEQ(sub, "all")) {
        char all[] = COMMAND_STAT_OPTIONS;

        command_stat(c, all, comm, NULL);
      } else {
        send_to_char(c, "Unknown stat option `%s`\n", sub);
        command_stat_usage(c);
      }
    }
  } else {
    command_stat_usage(c);
  }
#undef COMMAND_STAT_OPTIONS
}

void finger_item(character *c, const char *item, const char *string, ...)
{
  char buffer[OUTPUT_BUFFER];
  va_list ptr;

  va_start(ptr, string);
  mvsnprintf(buffer, sizeof(buffer), string, ptr);
  va_end(ptr);

  send_to_char(c, "%-10 10s : ^H%s^n\n", item, buffer);
} 


void command_finger(character *c, char *r, const char *comm, void *s)
{
  /* formats 
     finger       - does a finger on the character that called func
     finger bob   - does a finger on bob
     check for
       r = ""
       r logged in 
       r not logged in and exists
       r not logged in and doesn't exist
  */
  descriptor *des = getdes(c); /* used for bar2() funcs */
  descriptor *victdes;
  descriptor *load_des = pool_malloc(descriptor_pool);
  character *victim;
  char timebuffer[OUTPUT_BUFFER];
  bool loggedin = FALSE;
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */

  if (!load_des) {
    fprintf(stderr, "descriptor_new: Out of memory.");
    log(usage, "not enough memory in command_finger");
    send_to_char(c, "Panic: running out of memory\n");
    return;
  }

  if(STREQ(r, "")) {
    victim = c;
    victdes = getdes(victim);
    loggedin = TRUE;
  } else {
    if((victim = character_from_name(r)) != NULL) {
      /* victim is logged on and found */
      victdes = getdes(victim);
      loggedin = TRUE;
    } else {
      /* check to see is r is in the players directory, if so load
         else return */
      int failed;
      failed = load_player(load_des, r);
      if (failed) {
        send_to_char(c, "Player %s, not found\n", r);
        return;
      } else {
        victim = &load_des->player;
        victdes  = load_des;
        loggedin = FALSE;
      }
    }
  }

  if (loggedin) {
    time_t now = time(NULL);
    seconds_to_string(victim->prevtime + difftime(now, victim->logintime),
                    timebuffer, sizeof(timebuffer));
  } else {
    seconds_to_string(victim->prevtime, timebuffer, sizeof(timebuffer));
  }
  bar2(des, "Finger", NULL, NULL);
  finger_item(c, "Name", "%s", victim->name);
  finger_item(c, "Total Time", "%s", timebuffer);

  /* Display privileges */
/*  if (victim == c || haspriv(c, "s_base")) {
    send_to_char(c, "     Privs:");
    privs_list_player(c, victim);
    send_to_char(c, "^n\n");
  } */
  switch(victim->rank) {
    case 0:
      finger_item(c, "Rank", "Guest");
      break;
    case 1:
      finger_item(c, "Rank", "User");
      break;
    case 2:
      finger_item(c, "Rank", "Super User");
      break;
    case 3:
      finger_item(c, "Rank", "Admin");
      break;
    case 4:
      finger_item(c, "Rank", "Head Admin");
      break;
  }

  if (loggedin) {
    finger_item(c, "Group", "%s", victim->group);
    if(STREQ(victdes->termtype, "")) {
      finger_item(c, "Termtype", "(%ux%u)", victdes->term_width, 
                  victdes->term_height);
    } else {
      finger_item(c, "Termtype", "%s (%ux%u)", victdes->termtype,
                  victdes->term_width, victdes->term_height);
    }
    if (haspriv(c, "s_trace")) {
      char buf[128];
      format_host(buf, sizeof(buf), victdes);      
      finger_item(c, "Host", "%s", buf);
    }
  }
#ifdef USE_PLUGINS
  /* handle plugin extra data */
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    void (*function)(character *, character *, void *);

    if((p->onload != NULL) && (p->offset != NO_PLUGIN_EXTRA_DATA)) {
      function = p->onfinger;
      function(c, victim, ((char *)victdes) + p->offset);
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  bar2(des, NULL, NULL, NULL);

  pool_free(load_des);
}


void command_group(character *c, char *r, const char *comm, void *s)
{
  char groupname[MAX_GROUP_NAME_LENGTH+1];

  stripCodes(groupname, r);

  if(STREQ(r, "")) {
    STRNCOPY(groupname, "Public", sizeof(groupname));
  }
  if (strlen(r) > MAX_GROUP_NAME_LENGTH) {
    send_to_char(c, "Group name too long (max %d)\n", MAX_GROUP_NAME_LENGTH);
    return;
  }

  strlower(groupname);
  groupname[0] = toupper(groupname[0]);

  if (!STREQ(groupname, c->group)) {
    /* Announce departure of character from old group */
    send_to_room_except(c->group, c, "\n+++ %s joins group %s\n",
                        c->name, groupname);
    STRNCOPY(c->group, groupname, sizeof(c->group));
    /* annouce arrival in new group */
    send_to_char(c, "+++ You have joined group ^y%s^n\n", c->group);
    send_to_room_except(c->group, c, "\n+++ %s joins this group\n", c->name);
  } else {
    send_to_char(c, "You are already in group ^y%s^n\n", c->group);
  }
}

void command_dc(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = NULL;
  character *ch;

  if (STREQ(r, "")) {
    help_usage(c, "bump/dc");
    return;
  }

  ch = character_from_name(r);
  if (ch) {
    /* found character, drop em */
    des = getdes(ch);
  } else {
    /* char name not found - assume a number */
    if (!strisdigit(r)) {
      /* Not a number - report bad character name */
      send_to_char(c, "Character %s not known\n", r);
      return;
    } else {
      int sfd = atoi(r);
      listnode *dNode = AllConns.head.next;

      while (LIST_NODE_IS_REAL(dNode)) {
        descriptor *des2 = LIST_GET_DATA(dNode, descriptor *, descriptorlink);
        if (sfd == des2->sfd) {
          des = des2;
          break;
        }
        dNode = dNode->next;
      }
      if (!des) { 
        send_to_char(c, "SFD %d not found\n", sfd);
        return;
      }
    }
  }

  /* Before we drop, check we are not dropping ourselves */
  if (des == getdes(c)) {
    send_to_char(c, "Haven't you heard of the quit command?\n");
    return;
  }

  /* Check that the victim is not a higher rank than the caller */
  if(c->rank <= des->player.rank) {
    send_to_char(c, "Hmm, I don't think they'd like that, do you?\n");
    return;
  }

  des->state = STATE_CLOSING;
  if (des->player.loggedin) {
    send_to_all_except(&des->player, "\n+++ ^H%s^n has logged out\n",
                       des->player.name);
  }
}

void command_time(character *c, char *r, const char *comm, void *s)
{
  time_t now;
  struct tm nowtm;
  char buff[OUTPUT_BUFFER];
  unsigned long howlong;

  now = time(NULL);
  nowtm = *localtime(&now);
  strftime(buff, sizeof(buff), "%a %b %d %H:%M:%S %Y", &nowtm);
  send_to_char(c, "^WCurrent  : %s\n^n", buff);

  nowtm = *localtime(&starttime);
  strftime(buff, sizeof(buff), "%a %b %d %H:%M:%S %Y", &nowtm);
  send_to_char(c, "Up Since : %s\n", buff);

  howlong = (unsigned long) difftime(now, starttime);
  seconds_to_string(howlong, buff, sizeof(buff));
  send_to_char(c, "Up Time  : %s\n", buff);
}

void command_ping(character *c, char *r, const char *comm, void *s) 
{
  character *target;

  if (STREQ(r, "")) {
    help_usage(c, "ping");
  } else {  
    target = character_from_name(r);
    if (target == NULL) {
      send_to_char(c, "%s not found\n", r);
    } else if(target == c) {
      send_to_char(c, "Try pinging someone other than yourself\n", r);
    } else {
      send_to_char(target, "\a^HPING\n^n%s is trying to contact you\n",
                   c->name);
    }
  }
}

void command_setrank(character *c, char *r, const char *comm, void *s)
{
  character *target;
  char *rankname = r;
 
  if(STREQ(r, "")) {
    help_usage(c, "setrank");
    return;
  }

  strmunch(r, &r, &rankname);
  target = character_from_name(r);
  if (!target) {
    send_to_char(c, "Player %s not found\n", r);
    help_usage(c, "setrank");
    return;
  }
  if (c == target) {
    send_to_char(c, "You may not alter your own rank\n");
    return;
  }
  /* if rankname is blank print out usage and bail out */
  if (STREQ(rankname, "")) {
    send_to_char(c, "Rank name not found\n");
    help_usage(c, "setrank");
    return;
  }
  /* search through the priv list and try to find the priv */
  if (rankname_exists(rankname)) {
    send_to_char(c, "You have altered %s's rank to %s.\n",
                 target->name, rankname);
    send_to_char(target, "Your rank changes to %s\n",
                 rankname);
    setrank(target, rankname);
  } else {
    send_to_char(c, "Rank %s not found\n", rankname);
    help_usage(c, "setrank");
  }
}

void command_remove(character *c, char *r, const char *comm, void *s)
{
  if (STREQ(r, "")) {
    help_usage(c, "remove");
    return;
  }
  {
    character *target;
    char *privname = r;
   
    strmunch(r, &r, &privname);
    /* r should now be player name, privname should be the priv */
    /* try to find player, if not found bail out of function */
    target = character_from_name(r);
    if (!target) {
      send_to_char(c, "Player %s not found\n", r);
      help_usage(c, "remove");
      return;
    }
    /* check if removing privs from oneself */
    if (c == target) {
      send_to_char(c, "You may not remove your own privs\n");
      return;
    }
    /* if privname is blank print out usage and bail out */
    if (STREQ(privname, "")) {
      send_to_char(c, "Unknown priv\n");
      help_usage(c, "remove");
      return;
    }
    /* search through the priv list and try to find the priv */
    if (privname_exists(privname)) {
      send_to_char(c, "You have removed from %s the %s priv.\n",
                   target->name, privname);
      send_to_char(target, "You have lost the %s priv\n",
                   privname);
      clearpriv(target, privname);
    } else {
      send_to_char(c, "Priv %s not found\n", privname);
      help_usage(c, "remove");
    }
  }
}

static void command_grant_usage(character *c)
{
  help_usage(c, "grant");
  privs_list_all(c);
}

void command_grant(character *c, char *r, const char *comm, void *s) 
{
  if (STREQ(r, "")) {
    command_grant_usage(c);
    return;
  }
  {
    character *target;
    char *privname = r;
   
    strmunch(r, &r, &privname);
    /* r should now be player name, privname should be the priv */
    /* if privname is blank print out usage and bail out */
    if (STREQ(privname, "")) {
      command_grant_usage(c);
      return;
    }
    /* try to find player, if not found bail out of function */
    target = character_from_name(r);
    if (!target) {
      send_to_char(c, "Player %s not found\n", r);
      command_grant_usage(c);
      return;
    }
    /* search through the priv list and try to find the priv */
    if (privname_exists(privname)) {
      send_to_char(c, "You have granted %s the %s priv.\n",
                   target->name, privname);
      send_to_char(target, "You have been granted the %s priv\n",
                   privname);
      setpriv(target, privname);
    } else {
      send_to_char(c, "Priv %s not found\n", privname);
      command_grant_usage(c);
    }
  }
}

void command_lsd(character *c, char *r, const char *comm, void *s)
{
  listnode *dNode = AllConns.head.next;
  descriptor *des2;
  unsigned long howlong;
  time_t now;
  
  des2 = getdes(c);

  bar2(des2, "Sockets connected and states", NULL, NULL);
  send_to_char(c, "^H SFD State    (name)        (login)  Idle    Host^n\n\n");
  while (LIST_NODE_IS_REAL(dNode)) {
    descriptor *des = LIST_GET_DATA(dNode, descriptor *, descriptorlink);
    character *c2 = &des->player;
    char *state = "";

    send_to_char(c, "%3 3d ", des->sfd);
    switch (des->state) {
      case STATE_LOGIN:    state = "LOGIN";    break;
      case STATE_NEW1:     state = "NEW1";     break;
      case STATE_NEW2:     state = "NEW2";     break;
      case STATE_NEW3:     state = "NEW3";     break;
      case STATE_PASSWORD: state = "PASSWORD"; break;
      case STATE_PLAY:     state = "PLAY";     break;
      case STATE_CLOSING:  state = "CLOSING";  break;
    }
    send_to_char(c, "%-9s", state);
    now = time(NULL);
    if (des->state == STATE_PLAY) {
      send_to_char(c, "%-13s", c2->name);
      /* login Time */
      howlong = (unsigned long) difftime(now, c2->logintime);
      send_to_char(c, " %02lu:", howlong / 3600);
      howlong %= 3600;
      send_to_char(c, "%02lu:", howlong / 60);
      howlong %= 60;
      send_to_char(c, "%02lu", howlong);
    } else {
      send_to_char(c, "                      ");
    }
    /* Idle Time */
    howlong = (unsigned long) difftime(now, des->idletime);
    send_to_char(c, " %02lu:", howlong / 60);
    howlong %= 60;

    /* total time and hostname */
    {
      char buf[128];
      format_host(buf, sizeof(buf), des);
      send_to_char(c, "%02lu   %s\n", howlong, buf);
    }
    
    dNode = dNode->next;
  }
  bar2(des2, NULL, NULL, NULL);
}

void command_bug(character *c, char *r, const char *comm, void *s)
{
  if (STREQ(r, "")) {
    help_usage(c, comm);
  } else {
    if(STREQ(comm, "bug")) {
      log(bug, "%s - %s", c->name, r);
      send_to_char(c, "Thank you, the bug has been recorded\n");
    } else if(STREQ(comm, "idea")) {
      log(idea, "%s - %s", c->name, r);
      send_to_char(c, "Thank you, your idea has been recorded\n");
    } else if(STREQ(comm, "typo")) {
      log(typo, "%s - %s", c->name, r);
      send_to_char(c, "Thank you, the typo has been recorded\n");
    }
  }
}

void command_shutdown(character *c, char *r, const char *comm, void *s)
{
  sdown = TRUE;
}

void command_motd(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c);

  if(STREQ(comm, "motd")) {
    bar2(des, "Message Of The Day", NULL, NULL);
    send_file_to_descriptor("lib/etc/motd", des);
    bar2(des, NULL, NULL, NULL);
  } else if(STREQ(comm, "sumotd")) {
    bar2(des, "Super-user Message Of The Day", NULL, NULL);
    send_file_to_descriptor("lib/etc/sumotd", des);
    bar2(des, NULL, NULL, NULL);
  }
}

void command_prompt(character *c, char *r, const char *comm, void *s)
{
  if(strlen(r) > MAX_PROMPT_LENGTH-2) {
    send_to_char(c, "Prompt too long (max %d)\n", MAX_PROMPT_LENGTH-2);
  } else {
    STRNCOPY(c->prompt, r, sizeof(c->prompt));
  } 
}

void command_commands(character *c, char *r, const char *comm, void *s)
{
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */

  descriptor *des = getdes(c);
  int l;

  bar2(des, "List of commands", NULL, NULL);
  send_to_char(c, "^g[^n^Hbase^N^g]^n\n");
  for(l=0; l<commandn; l++) {
    const command *co = &commands[l];
    if(haspriv(c, co->privname)) {
      send_to_char(c, "%s%s^n ", privcol(co->privname),
                   co->name); 
    }
  }
  send_to_char(c, "\n");

#ifdef USE_PLUGINS
  while(LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    if(p->clist != NULL) {
      int num = 0;
      command *co = p->clist;
      /* count the num of commands visible to this character */
      while(co->name != NULL) {
        if(haspriv(c, co->privname))
          num++;
        co++;
      }
      if(num != 0) {
        co = p->clist;
        send_to_char(c, "^g[^n^H%s^N^g]^n\n", p->name);
        while(co->name != NULL) {
          if(haspriv(c, co->privname))
            send_to_char(c, "%s%s^n ", privcol(co->privname), co->name);
          co++;
        }
        send_to_char(c, "\n");
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  bar2(des, NULL, NULL, NULL);
}

void command_users(character *c, char *r, const char *comm, void *s)
{
  listnode *pNode = AllPlayers.head.next;

  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode,character *, characterlink);
    send_to_char(c, "%s ", pChar->name);
    pNode = pNode->next;
  }
  send_to_char(c, "\n");
}

#define FLAG_LEN 4
void command_who(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c);

  listnode *pNode = AllPlayers.head.next;
  int howmany = 0;
  time_t now;
  unsigned long howlong;
  char flags[20] = "U";

  now = time(NULL);
  bar2(des, "Who is logged on?", NULL, NULL);
  send_to_char(c, "^H%-*s%-*s Login    Idle    %-*s ",
               MAX_NAME_LENGTH, "Name", FLAG_LEN, "Flg", MAX_GROUP_NAME_LENGTH, "Group");
  if(haspriv(c, "s_trace")) {
    send_to_char(c, "Host");
  }
  send_to_char(c, "^n\n");
  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode, character *, characterlink);
    descriptor *des = LIST_GET_DATA(pChar, descriptor *, player);

    /* Display name */
    send_to_char(c, "%-*s", MAX_NAME_LENGTH, pChar->name);
    /* Do Flags */
    switch(pChar->rank) {
      /* oh for some #define's */
      case 1:
        STRNCOPY(flags, "U", sizeof(flags));
        break;
      case 2:
        STRNCOPY(flags, "S", sizeof(flags));
        break;
      case 3:
        STRNCOPY(flags, "A", sizeof(flags));
        break;
      case 4:
        STRNCOPY(flags, "H", sizeof(flags));
        break;     
    }
    /* handle group owner flag, when new groups */
    send_to_char(c, "%-*s", FLAG_LEN, flags);

    /* Calculate how long they have been logged in */
    howlong = (unsigned long) difftime(now, pChar->logintime);
    send_to_char(c, " %02lu:", howlong / 3600);
    howlong %= 3600;
    send_to_char(c, "%02lu:", howlong / 60);
    howlong %= 60;
    send_to_char(c, "%02lu", howlong);
    /* Calculate how long they have been idle for */
    howlong = (unsigned long) difftime(now, des->idletime);
    send_to_char(c, " %02lu:", howlong / 60);
    howlong %= 60;
    send_to_char(c, "%02lu   ", howlong);
    send_to_char(c, "%-*s ", MAX_GROUP_NAME_LENGTH, pChar->group);
    if(haspriv(c, "s_trace")) {
      char buf[128];
      format_host(buf, sizeof(buf), des);
      send_to_char(c, "%s", buf);
    }
    send_to_char(c, "\n");
    pNode = pNode->next;
    howmany++;
  }
  send_to_char(c, "Total: %d user%s\n", howmany, howmany != 1 ? "s" : "");
  send_to_char(c, "(U) = User, (S) = Super User, (A) = Admin, (H) = Head Admin\n");
  bar2(des, NULL, NULL, NULL);
}

void command_tell(character *c, char *sRemainder, const char *comm, void *s)
{
  char *sMessage;
  character *pDest;

  strmunch(sRemainder, &sRemainder, &sMessage);

  pDest = character_from_name(sRemainder);

  if (!pDest) {
    send_to_char(c, "Character '%s' could not be found or is ambiguous.\n", sRemainder);
    return;
  }

  if (pDest == c) {
    send_to_char(c, "Talking to yourself again?\n");
    return;
  }

  if(STREQ(sMessage, "")) {
    help_usage(c, comm);
    return;
  }

  if(STREQ(comm, "tell")) {  
    do_tell(pDest, "%s tells you '%s^H'\n", c->name, sMessage);
    send_to_char(c, "You tell %s '%s^n'\n", pDest->name, sMessage);
  } else if(STREQ(comm, "remote")) {
    do_tell(pDest, "%s %s, at you\n", c->name, sMessage);
    send_to_char(c, "You emote '%s %s^n' to %s^n\n", c->name, sMessage,
                 pDest->name);
  }
}

void command_say(character *c, char *r, const char *comm, void *s)
{
  if (STREQ(r, "")) {
    help_usage(c, comm);
    return;
  }
  if(STREQ(comm, "say")) {
    send_to_room_except(c->group, c, "%s says \'%s^n\'\n", c->name, r);
    send_to_char(c, "You say \'%s^n\'\n", r);
  } else if(STREQ(comm, "emote")) {
    send_to_room_except(c->group, c, "%s %s\n", c->name, r);
    send_to_char(c, "You emote \'%s %s\'\n", c->name, r);
  }
}

void command_shout(character *c, char *r, const char *comm, void *s)
{
  if (STREQ(r, "")) {
    help_usage(c, "shout");
    return;
  }
  send_to_all_except(c, "%s shouts \'%s^n\'\n", c->name, r);
  send_to_char(c, "You shout \'%s^n\'\n", r);
}

void command_quit(character *c, char *r, const char *comm, void *s)
{
  descriptor *des = getdes(c); 

  des->state = STATE_CLOSING;
  if(r && !STREQ(r, "")) {
    send_to_all_except(c, "\n+++ ^r%s^n has logged out (%s)\n", c->name, r);
  } else {
    send_to_all_except(c, "\n+++ ^r%s^n has logged out\n", c->name);
  }
  send_to_char(c, "\n+++ You have logged out\n");
  if (send_file_to_descriptor("lib/etc/logout", des) !=0) {
    send_to_descriptor(des, "Goodbye!\n");
  }

  des->safe = FALSE;
}

void command_whoami(character *c, char *r, const char *comm, void *s)
{
  send_to_char(c, "You are ^r%s^n\n", c->name);
}

bool command_shortcut_run(descriptor *des, char *s)
{
  listnode *cNode = AllCommands.head.next;
  character *p = &des->player;

  s = strip_leading_white_space(s);
  if (strlen(s) != 0) {
    while(LIST_NODE_IS_REAL(cNode)) {
      command_list_item *cli=LIST_GET_DATA(cNode, command_list_item *, commandlink); 
      commandl *c = &cli->c;
      if(strchr(c->shortcut, s[0]) && haspriv(p, c->privname)) {
        void (*function)(character *, char *, const char *, void *);

        function = c->function;
        s++; /* to get past the shortcut character */
        s = strip_leading_white_space(s);
        if(c->offset == NO_PLUGIN_EXTRA_DATA) {
          function(p, s, c->name, NULL);
        } else {
           /* calculate offset to plugin player data, and call */
           descriptor *d = getdes(p);
           function(p, s, c->name, ((char *)d) + c->offset);
        }
       return TRUE;
      }
      cNode = cNode->next;
    }
  }

  return FALSE;
}

void command_parse_n_run(descriptor *des, const char *first,
                         char *remainder)
{
  listnode *cNode = AllCommands.head.next;
  commandl *remember = NULL; /* if partial match remember */
  unsigned int count = 0;
  character *p = &des->player;

  if (STREQ(first, ""))
    return;

  while(LIST_NODE_IS_REAL(cNode)) {
    command_list_item *cli=LIST_GET_DATA(cNode, command_list_item *, commandlink); 
    commandl *c = &cli->c;

    if(STREQ(first, c->name) && haspriv(p, c->privname))
    {
      /* Matches exactly */
      void (*function)(character *, char *, const char *, void *);

      function = c->function;
      if(c->offset == NO_PLUGIN_EXTRA_DATA) {
        function(p, remainder, c->name, NULL);
      } else {
        /* calculate offset to plugin player data, and call */
        descriptor *d = getdes(p);
        function(p, remainder, c->name, ((char *)d) + c->offset);
      }
      return;
    } else if (STRNEQ(first, c->name, strlen(first))
               && haspriv(p, c->privname)) {
      /* Part matches */
      remember = c;
      count ++;
    }
    cNode = cNode->next;
  }
  switch (count) {
    case 0:
      send_to_char(p, "Command \'%s\' not known, try help\n", first);
      break;
    case 1:
      /* Matches one command without ambiguity */
      {
        void (*function)(character *, char *, const char *, void *);

        function = remember->function;
        if(remember->offset == NO_PLUGIN_EXTRA_DATA) {
          function(p, remainder, remember->name, NULL);
        } else {
          /* calculate offset to plugin player data, and call */
          descriptor *d = getdes(p);
          function(p, remainder, remember->name, ((char *)d) + remember->offset);
        }
        return;
        break;
      }
    default:
      /* Part matches more than one command */
      send_to_char(p, "Command \'%s\' ambiguous, try help\n", first);
      break;
   }
}

void command_do(descriptor *des, const char *buffer)
{
  char line[MAX_RAW_INPUT_BUFFER];
  char *first, *remainder;
  char *temp;
  bool carryon = TRUE;

  STRNCOPY(line, buffer, sizeof(line));
  first = line;

  if(haspref(&des->player, "nuts")) {
    /* NUTS style */
    first = strip_leading_white_space(first);
    strip_trailing_white_space(first);  
    if(*first == '.') {
      first++; /* skip past the dot */
      if (command_shortcut_run(des, first) == FALSE) {
        strmunch(first, &first, &remainder);
        strlower(first);
        command_parse_n_run(des, first, remainder);
      }
    } else {
      command_say(&des->player, first, "say", NULL);
    }
  } else {
    /* EW-TOO style */
    while (carryon == TRUE) {
      /* strip leading white space in command */
      first = strip_leading_white_space(first);
      /* search for a %; in the input */
      temp = strstr(first+1, "%;");
      if (temp != NULL) {
        /* found a ; */
        *temp = '\0';
        temp+= 2;
        if (*temp == '\0') {
          /* is temp a 0 length string ? */
          carryon = FALSE;
        } else {
          carryon = TRUE;
        }
      } else {
        carryon = FALSE;
      }

      strip_trailing_white_space(first);  
      if (command_shortcut_run(des, first) == FALSE) {
        strmunch(first, &first, &remainder);
        strlower(first);
        command_parse_n_run(des, first, remainder);
      }

      if (carryon == TRUE) {
        first = temp;
      }
    }
  }

  /* Do a prompt */
  send_to_char(&des->player, "%s^n", &des->player.prompt);
}

descriptor* getdes(character *c) {
  return LIST_GET_DATA(c, descriptor *, player);
}

void seconds_to_string(unsigned long seconds, char *buffer, size_t size)
{
  unsigned long t;
  int len;

  t = seconds / (60 * 60 * 24 * 365);
  if (t > 0) {
    len = msnprintf(buffer, size, "%lu year%s, ", t, t != 1 ? "s" : "");
    buffer += len;
    size -= len;
  }
  seconds %= (60 * 60 * 24 * 365);
  t = seconds / (60 * 60 * 24);
  if (t > 0) {
    len = msnprintf(buffer, size, "%lu day%s, ", t, t != 1 ? "s" : "");
    buffer += len;
    size -= len;
  }
  seconds %= (60 * 60 * 24);
  t = seconds / (60 * 60);
  if (t > 0) {
    len = msnprintf(buffer, size, "%lu hr%s, ", t, t != 1 ? "s" : "");
    buffer += len;
    size -= len;
  }
  seconds %= (60 * 60);
  t = seconds / 60;
  if (t > 0) {
    len = msnprintf(buffer, size, "%lu min%s, ", t, t != 1 ? "s" : "");
    buffer += len;
    size -= len;
  }
  seconds %= 60;
  msnprintf(buffer, size, "%lu sec%s", seconds, seconds != 1 ? "s": "");
}
