/*
  Calisto - An internet Talker
  Copyright (C) 1998-2001 Peter Howkins, Matthew Howkins and Simon Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,     
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License   
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: calisto@calisto.cx

  $Id: help.c,v 1.14 2002/12/02 19:34:45 peter Exp $

 */
static const char rcsid[] = "$Id: help.c,v 1.14 2002/12/02 19:34:45 peter Exp $";

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "help.h"
#include "library.h"
#include "log.h"
#include "mem.h"
#include "msnprintf.h"
#ifdef USE_PLUGINS
 #include "plugin.h"
 #include "plugin2.h"
#endif /* USE_PLUGINS */
#include "strplus.h"
#include "structs.h"

typedef struct {
  const char *entry;
  const char *text;
} help_item;

static char *helptext = NULL;
static help_item *help_items;
static unsigned help_item_count;
static const char *help_path = NULL;

static char *getlinefromhelp(const char *text, char *buffer)
{
  while (*text && *text != 10 && *text != 13)
    *buffer++ = *text++;

  if (*text == 10) {
    text++;
    if (*text == 13)
      text++;
  } else if (*text == 13) {
    text++;
    if (*text == 10)
      text++;
  }

  *buffer = '\0';
  return (char *) text;
}

static unsigned buffer_word_count(const char *buffer)
{
  unsigned wc = 0;

  while (*buffer) {
    while (*buffer && isspace(*buffer))
      buffer++;
    if (*buffer) {
      wc++;
      while (*buffer && !isspace(*buffer))
        buffer++;
    }
  }
  return wc;
}

static int help_process(void)
{
/*  int hack =0;
*/
  char *t = helptext;
  char buffer[1024];
  bool ignoring = FALSE; /* whether we are ignoring lines at the moment, searching for lines
                            of interest */
  help_item *x;

  /* Count the number of helptext entries */
  help_item_count = 0;
  while (*t) {
    t = getlinefromhelp(t, buffer);
/*    log(debug, "%s", buffer);*/
    if (!ignoring) {
      /* Count the number of words in the buffer at the moment */
      help_item_count += buffer_word_count(buffer);
      ignoring = TRUE;
    } else {
      /* Ignore lines until we find a line that is just a '#' on its own */
      if (STREQ(buffer, "#"))
        ignoring = FALSE;
    }
  }
/*  log(debug, "Help: Helptext entry count is %u", help_item_count);*/

  /* Allocate space for array of helptext entries */
  help_items = MEM_MALLOC(help_item_count * sizeof(help_item));
  if (!help_items) {
    log(usage, "Help: Not enough memory for help file index - no help available");
    helptext = NULL;
    return 1;
  }

  /* Parse the helptext again, changing '\n' to '\0', putting '\0' between index words,
     and filling in the index */
  t = helptext;
  ignoring = FALSE;
  x = help_items;
  while (*t) {
    char *newline = strchr(t, '\n');
    char *nextline;

    if (newline) {
      /* flibble hack to remove /r's from the end of lines that was causing seg
         faults */
/*      if(*(newline - 1) == '\r') {
        *(newline - 1) = '\0';
        *newline = '\0';
        nextline = newline + 1;
        hack =1;
      } else { 
*/
        *newline = '\0';
        nextline = newline + 1;
/*      } 
*/
    } else {
      nextline = t + strlen(t);
    }

    if (!ignoring) {
      /* Index the words from t at the moment */
      while (*t) {
        while (*t && isspace(*t))
          t++;
        if (*t) {
          x->entry = t;
          x->text  = nextline;
          x++;
          while (*t && !isspace(*t))
            t++;
          if (*t)
            *t++ = '\0';
        }
      }
      ignoring = TRUE;
    } else {
      if (STREQ(t, "#")) {
        ignoring = FALSE;
        *t = '\0';
      } else {
        if (newline) {
/*          if(hack) {
            *(newline-1) = '\n';
            *newline = ' ';
            hack =0;
          } else {
*/
            *newline = '\n';
/*          }
*/
        }
      }
    }

    t = nextline;
  }
  log(debug, "sild");
  return 0;
}

int help_item_compare(const void *a, const void *b)
{
  const help_item *c = a;
  const help_item *d = b;

  assert(a);
  assert(b);
  assert(c);
  assert(d);
  assert(c->entry);
  assert(d->entry);

  return stricmp(c->entry, d->entry);
}

/* Sort the helptext entries, ready for binary searching later */
static void help_sort(void)
{
  qsort(help_items, (size_t) help_item_count, sizeof(help_item), help_item_compare);
}

void help_init(const char *path)
{
  FILE *f;
  unsigned helpfilesize, mainhelpsize;
  char *helploop;
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */

  help_path = path;
  /* Open help file to determine its presence and size */
  f = fopen(help_path, "rb");
  if (!f) {
    log(usage, "Help: Unable to open file '%s' - no help availble", help_path);
    return;
  }
  (void) fseek(f, 0, SEEK_END);
  mainhelpsize = (unsigned) ftell(f);
  helpfilesize = mainhelpsize;

#ifdef USE_PLUGINS
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    if(p->helpfile && !STREQ(p->helpfile, "")) {
      FILE *f2;
      char path[256];
      msnprintf(path, sizeof(path), "%s%s", PLUGINDIR, p->helpfile);

      f2 = fopen(path, "rb");
      if(f2) {        
        (void) fseek(f2, 0, SEEK_END);
        helpfilesize += (unsigned) ftell(f2);

        fclose(f2);
      } else {
        fprintf(stderr, "Error Loading '%s' helpfile\n", path);
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

/*  log(debug, "Help: Help file size is %u", helpfilesize);*/

  /* Allocate memory for, and load in the help file */
  helptext = MEM_MALLOC((size_t) helpfilesize + 1);
  if (!helptext) {
    log(usage, "Help: Not enough memory for help file - no help available");
    fclose(f);
    return;
  }
  rewind(f);
  helploop = helptext;
  if (fread(helploop, 1, (size_t) mainhelpsize, f) != (size_t) mainhelpsize) {
    log(usage, "Help: Problem loading help file - help not available");
    fclose(f);
    helptext = NULL;
    return;
  }
  fclose(f);
  

#ifdef USE_PLUGINS
  helploop += mainhelpsize;
  /* loop through freading plugin help files */
  pNode = AllPlugins.head.next;
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    unsigned filesize;
    if(p->helpfile && !STREQ(p->helpfile, "")) {
      FILE *f2;
      char path[256];
      msnprintf(path, sizeof(path), "%s%s", PLUGINDIR, p->helpfile);

      f2 = fopen(path, "rb");
      if(f2) {
        (void) fseek(f2, 0, SEEK_END);
        filesize = (unsigned) ftell(f2);
        rewind(f2);
        if (fread(helploop, 1, (size_t) filesize, f2) != (size_t) filesize) {
          log(usage, "Help: Problem loading help file '%s'- help not available", p->helpfile);
          fclose(f2);
          helptext = NULL;
          return;
        }
        helploop += filesize;

        fclose(f2);
      } else {
        fprintf(stderr, "Error Loading '%s' helpfile\n", path);
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */



  /* Add a zero terminator to the help text */
  helptext[helpfilesize] = '\0';

  log(usage, "Help: Successfully loaded '%s'", path);

  if (help_process())
    return;
  help_sort();

  /* Log the helptext index to the debug LOG */
/*  {
    unsigned item;

    for (item = 0; item < help_item_count; item++) {
      log(debug, "Help: Item % 2u : %s", item, help_items[item].entry);
      log(debug, "Help:   %s", help_items[item].text);
    }
  }*/
}

/* Process the string from src into dest.
   Look for fields starting with a $, and process appropriately. */
static void help_process_text(char *dest, size_t size, const char *src,
                              const character *c, const char *help_item)
{
  while (*src) {
    if (*src == '$') {
      src++;
      if (STRINEQ(src, "comm;", 5)) {
        src += 5;
        dest += msnprintf(dest, size, "%s", help_item);
      } else if (STRINEQ(src, "user;", 5)) {
        src += 5;
        dest += msnprintf(dest, size, "%s", c->name);
      } else {
        /* Leave the token unchanged, and copy it over.
           We only need the dollar, the rest will arrive later */
        *dest++ = '$';
      }
    } else {
      *dest++ = *src++;
    }
  }
  *dest = '\0';
}

void command_help(character *c, char *r, const char *comm, void *s)
{
  if (!helptext) {
    send_to_char(c, "There was a problem loading helptext during startup.\n"
                    "Help is therefore not available. Try moaning at an Admin\n");
    return;
  }

  if (*r == '\0') {
    unsigned i;

    /* They typed just help, therefore we list availble help entries */
    send_to_char(c, "Usage: help <help entry>\n"
                    "where <help entry> is one of the following:\n");
    for (i = 0; i < help_item_count; i++) {
      char buffer[256];
      STRNCOPY(buffer, help_items[i].entry, sizeof(buffer));
      strlower(buffer);
      send_to_char(c, "%s ", buffer);
    }
    send_to_char(c, "\nAlso, try typing \'commands\' to list available commands\n");
  } else {
    const help_item *item;
    help_item key;

    key.entry = r;
    item = bsearch(&key, help_items, (size_t) help_item_count,
                   sizeof(help_item), help_item_compare);
    if (item) {
      const char *t = item->text;
      char buffer[1024], buffer2[1024];

      while (*t) {
        char *b = buffer;

        t = getlinefromhelp(t, b);
        if (*b == '#') {
          b++;
          if (STRINEQ(b, "usage", 5)) {
            b += 5;
            while (*b && isspace(*b))
              b++;
            help_process_text(buffer2, sizeof(buffer2), b, c, r);
            send_to_char(c, "^YUsage: ^W%s^n\n", buffer2);
          } else if (STRINEQ(b, "seealso", 7)) {
            b += 7;
            while (*b && isspace(*b))
              b++;
            help_process_text(buffer2, sizeof(buffer2), b, c, r);
            send_to_char(c, "^YSee also: ^W%s^n\n", buffer2);
          }
        } else {
          help_process_text(buffer2, sizeof(buffer2), b, c, r);
          send_to_char(c, "%s\n", buffer2);
        }
      }
    } else
      send_to_char(c, "Help on %s not found\n", r);
  }
}

void help_reload(character *c)
{
  FILE *f;
  unsigned helpfilesize, mainfilesize;
  void *ok;
  char *helploop;
#ifdef USE_PLUGINS
  listnode *pNode = AllPlugins.head.next;
#endif /* USE_PLUGINS */

  /* Open help file to determine its presence and size */
  f = fopen(help_path, "rb");
  if (!f) {
    send_to_char(c, "Unable to open file '%s' - leaving help unchanged\n",
                 help_path);
    return;
  }
  (void) fseek(f, 0, SEEK_END);
  helpfilesize = (unsigned) ftell(f);
  mainfilesize = helpfilesize;

  helploop = helptext;
#ifdef USE_PLUGINS
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    if(p->helpfile && !STREQ(p->helpfile, "")) {
      FILE *f2;
      char path[256];
      msnprintf(path, sizeof(path), "%s%s", PLUGINDIR, p->helpfile);

      f2 = fopen(path, "rb");
      if(f2) {        
        (void) fseek(f2, 0, SEEK_END);
        helpfilesize += (unsigned) ftell(f2);

        fclose(f2);
      } else {
        send_to_char(c, "Error Loading '%s' helpfile\n", path);
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  /* Allocate memory for, and load in the help file */
  ok = realloc(helptext, (size_t) helpfilesize + 1);
  if (!ok) {
    send_to_char(c, "Not enough memory for help file - leaving help unchanged");
    fclose(f);
    return;
  }
  helptext = ok;
  rewind(f);
  if (fread(helptext, 1, (size_t) mainfilesize, f) != (size_t) mainfilesize) {
    send_to_char(c, "Problem loading help file - help not available");
    fclose(f);
    MEM_FREE(helptext);
    helptext = NULL;
    return;
  }
  fclose(f);

#ifdef USE_PLUGINS
  helploop += mainfilesize;
  /* loop through freading plugin help files */
  pNode = AllPlugins.head.next;
  while (LIST_NODE_IS_REAL(pNode)) {
    struct plugin *p = LIST_GET_DATA(pNode, struct plugin *, pluginlink);
    unsigned filesize;
    if(p->helpfile && !STREQ(p->helpfile, "")) {
      FILE *f2;
      char path[256];
      msnprintf(path, sizeof(path), "%s%s", PLUGINDIR, p->helpfile);

      f2 = fopen(path, "rb");
      if(f2) {
        (void) fseek(f2, 0, SEEK_END);
        filesize = (unsigned) ftell(f2);
        rewind(f2);
        if (fread(helploop, 1, (size_t) filesize, f2) != (size_t) filesize) {
          log(usage, "Help: Problem loading help file '%s'- help not available", p->helpfile);
          send_to_char(c, "Help: Problem loading help file '%s'- help not available", p->helpfile);
          fclose(f2);
          helptext = NULL;
          return;
        }
        helploop += filesize;

        fclose(f2);
      } else {
        send_to_char(c, "Error Loading '%s' helpfile\n", path);
      }
    }
    pNode = pNode->next;
  }
#endif /* USE_PLUGINS */

  /* Add a zero terminator to the help text */
  helptext[helpfilesize] = '\0';

  if (help_process())
    return;
  help_sort();
}

void help_usage(character *c, const char *command)
{
  const help_item *item;
  help_item key;

  /* Check for helptext being unavailable */
  if (!helptext)
    return;

  key.entry = command;
  item = bsearch(&key, help_items, (size_t) help_item_count,
                 sizeof(help_item), help_item_compare);
  /* Check for help item not found */
  if (!item)
    return;

  {
    const char *t = item->text;
    char buffer[1024], buffer2[1024];

    while (*t) {
      char *b = buffer;

      t = getlinefromhelp(t, b);
      if (*b == '#') {
        b++;
        if (STRINEQ(b, "usage", 5)) {
          b += 5;
          while (*b && isspace(*b))
            b++;
          help_process_text(buffer2, sizeof(buffer2), b, c, command);
          send_to_char(c, "Usage: %s\n", buffer2);
        }
      }
    }
  }
}

void help_close(void)
{
  MEM_FREE(helptext);
  MEM_FREE(help_items);
}
